# authentic2 - versatile identity manager
# Copyright (C) 2010-2022 Entr'ouvert
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import re
from unittest import mock

import lasso
import pytest
from mellon.adapters import UserCreationError

from authentic2.custom_user.models import User
from authentic2_auth_saml.adapters import MappingError
from authentic2_auth_saml.models import AddRoleAction, SAMLAuthenticator, SetAttributeAction


def test_lookup_user_ok(adapter, idp, saml_attributes, title_attribute):
    assert User.objects.count() == 0

    user = adapter.lookup_user(idp, saml_attributes)
    user.refresh_from_db()
    assert user.email == 'john.doe@example.com'
    assert user.attributes.title == 'Mr.'
    assert user.first_name == 'John'
    assert user.attributes.title == 'Mr.'
    assert user.ou.default is True


def test_lookup_user_missing_mandatory_attribute(adapter, idp, saml_attributes, title_attribute):
    del saml_attributes['mail']

    assert User.objects.count() == 0
    assert adapter.lookup_user(idp, saml_attributes) is None
    assert User.objects.count() == 0


def test_apply_attribute_mapping_missing_attribute_logged(
    caplog, adapter, idp, saml_attributes, title_attribute, user
):
    caplog.set_level('WARNING')
    saml_attributes['http://nice/attribute/givenName'] = []
    adapter.provision_a2_attributes(user, idp, saml_attributes)
    assert re.match('.*no value.*first_name', caplog.records[-1].message)


@pytest.mark.parametrize('action_name', ['add-role', 'toggle-role'])
class TestAddRole:
    @pytest.fixture
    def idp(self, action_name, simple_role):
        authenticator = SAMLAuthenticator.objects.create(
            enabled=True,
            metadata='meta1.xml',
            slug='idp1',
        )
        AddRoleAction.objects.create(authenticator=authenticator, role=simple_role)
        return authenticator.settings

    @pytest.fixture
    def saml_attributes(self):
        return {
            'issuer': 'https://idp.com/',
            'name_id_content': 'xxx',
            'name_id_format': lasso.SAML2_NAME_IDENTIFIER_FORMAT_PERSISTENT,
        }

    def test_lookup_user_success(self, adapter, simple_role, idp, saml_attributes):
        user = adapter.lookup_user(idp, saml_attributes)
        assert simple_role in user.roles.all()


def test_apply_attribute_mapping_missing_attribute_exception(
    adapter, idp, saml_attributes, title_attribute, user, rf
):
    saml_attributes['http://nice/attribute/givenName'] = []
    SetAttributeAction.objects.filter(user_field='first_name').update(mandatory=True)
    with pytest.raises(MappingError, match='no value'):
        adapter.provision_a2_attributes(user, idp, saml_attributes)

    request = rf.get('/')
    request._messages = mock.Mock()
    adapter.request = request
    with pytest.raises(UserCreationError):
        adapter.finish_create_user(idp, saml_attributes, user)
    request._messages.add.assert_called_once_with(
        40, 'User creation failed: no value for attribute "first_name".', ''
    )
