# authentic2 - versatile identity manager
# Copyright (C) 2010-2020 Entr'ouvert
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import pytest
from django.core.exceptions import ValidationError

from authentic2.forms.fields import PhoneField
from authentic2.forms.passwords import NewPasswordField


def test_phonenumber_field(settings):
    settings.DEFAULT_COUNTRY_CODE = '33'
    field = PhoneField()

    positive = [
        {'input': ['33', '01 01 01 01 01'], 'output': '+33101010101'},
        # undialable numbers are still parsed and usable as identifiers
        {'input': ['33', '01 01 01'], 'output': '+33010101'},
        {'input': ['33', '01 01 01 010101'], 'output': '+3310101010101'},
    ]
    # positive
    for value in positive:
        output = field.clean(value['input'])
        assert output == value['output']

    # negative
    for value in [
        ['33', '01a01'],
        ['33', '+01 01 01 01 01'],
        ['33', ' + 01/01.01-01.01'],
        ['33', '+01/01.01-01.01'],
    ]:
        with pytest.raises(ValidationError):
            field.clean(value)


def test_validate_password(settings):
    field = NewPasswordField()
    with pytest.raises(ValidationError):
        field.validate('aaaaaZZZZZZ')
    with pytest.raises(ValidationError):
        field.validate('00000aaaaaa')
    with pytest.raises(ValidationError):
        field.validate('00000ZZZZZZ')
    field.validate('000aaaaZZZZ')


@pytest.mark.parametrize(
    'password,min_strength',
    [
        ('?', 0),
        ('?JR!', 1),
        ('?JR!p4A', 2),
        ('?JR!p4A2i', 3),
        ('?JR!p4A2i:#', 4),
    ],
)
def test_validate_password_strength(settings, password, min_strength):
    settings.A2_PASSWORD_POLICY_MIN_LENGTH = len(password)
    field = NewPasswordField()

    field.min_strength = min_strength
    field.validate(password)

    settings.A2_PASSWORD_POLICY_MIN_LENGTH = len(password) + 1
    with pytest.raises(ValidationError):
        field.validate(password)

    if min_strength < 4:
        settings.A2_PASSWORD_POLICY_MIN_LENGTH = len(password)
        field.min_strength = min_strength + 1
        with pytest.raises(ValidationError):
            field.validate(password)


def test_digits_password_policy(settings):
    settings.A2_PASSWORD_POLICY_REGEX = '^[0-9]{8}$'
    settings.A2_PASSWORD_POLICY_REGEX_ERROR_MSG = 'pasbon'
    settings.A2_PASSWORD_POLICY_MIN_LENGTH = 0
    settings.A2_PASSWORD_POLICY_MIN_CLASSES = 0
    field = NewPasswordField()

    with pytest.raises(ValidationError):
        field.validate('aaa')
    field.validate('12345678')
